/*
 *  include/asm-st200/atomic.h
 *
 *  Copyright (C) 2002 STMicroelectronics Limited
 *	Author: Stuart Menefy <stuart.menefy@st.com>
 *	        Dave Mckay <david.mckay@st.com>
 *
 */

#ifndef _ASM_ST200_ATOMIC_H
#define _ASM_ST200_ATOMIC_H

#include <linux/types.h>

/*
 * Atomic operations that C can't guarantee us.  Useful for
 * resource counting etc..
 *
 */

#define ATOMIC_INIT(i)	{ (i) }

/**
 * atomic_read - read atomic variable
 * @v: pointer of type atomic_t
 *
 * Atomically reads the value of @v.
 */
#define atomic_read(v)		((v)->counter)

/**
 * atomic_set - set atomic variable
 * @v: pointer of type atomic_t
 * @i: required value
 *
 * Atomically sets the value of @v to @i.
 */
#define atomic_set(v,i)		((v)->counter = (i))

/**
 * atomic_add - add integer to atomic variable
 * @i: integer value to add
 * @v: pointer of type atomic_t
 *
 * Atomically adds @i to @v.
 */
static inline void atomic_add(int i, atomic_t * v)
{
	int tmp;

	asm volatile ("1 : ldw %0   = 0[%1] \n"
		      "    mov $r0.49 = 1b \n"
		      "    or  $r0.1 = $r0.1,1 \n"
		      "    ;; \n"
		      "    add %0 = %0,%2 \n"
		      "    ;; \n"
		      "    stw 0[%1] = %0 \n"
		      "    and $r0.1 = $r0.1,~1\n":"=&r" (tmp)
		      :"r"(v), "r"(i)
		      :"r0.49", "memory");
}

/**
 * atomic_sub - subtract the atomic variable
 * @i: integer value to subtract
 * @v: pointer of type atomic_t
 *
 * Atomically subtracts @i from @v.
 */
static inline void atomic_sub(int i, atomic_t * v)
{
	int tmp;

	asm volatile ("1 : ldw %0   = 0[%1] \n"
		      "    mov $r0.49 = 1b \n"
		      "    or  $r0.1 = $r0.1,1 \n"
		      "    ;; \n"
		      "    sub %0 = %0,%2 \n"
		      "    ;; \n"
		      "    stw 0[%1] = %0 \n"
		      "    and $r0.1 = $r0.1,~1\n":"=&r" (tmp)
		      :"r"(v), "r"(i)
		      :"r0.49", "memory");
}

static inline int atomic_add_return(int i, atomic_t * v)
{
	int tmp;

	asm volatile ("1 : ldw %0   = 0[%1] \n"
		      "    mov $r0.49 = 1b \n"
		      "    or  $r0.1 = $r0.1,1 \n"
		      "    ;; \n"
		      "    add %0 = %0,%2 \n"
		      "    ;; \n"
		      "    stw 0[%1] = %0 \n"
		      "    and $r0.1 = $r0.1,~1\n":"=&r" (tmp)
		      :"r"(v), "r"(i)
		      :"r0.49", "memory");
	return tmp;
}

static inline int atomic_sub_return(int i, atomic_t * v)
{
	int tmp;

	asm volatile ("1 : ldw %0   = 0[%1] \n"
		      "    mov $r0.49 = 1b \n"
		      "    or  $r0.1 = $r0.1,1 \n"
		      "    ;; \n"
		      "    sub %0 = %0,%2 \n"
		      "    ;; \n"
		      "    stw 0[%1] = %0 \n"
		      "    and $r0.1 = $r0.1,~1\n":"=&r" (tmp)
		      :"r"(v), "r"(i)
		      :"r0.49", "memory");

	return tmp;
}

static inline int atomic_inc_return(atomic_t * v)
{
	int tmp;

	asm volatile ("1 : ldw %0   = 0[%1] \n"
		      "    mov $r0.49 = 1b \n"
		      "    or  $r0.1 = $r0.1,1 \n"
		      "    ;; \n"
		      "    add %0 = %0,1 \n"
		      "    ;; \n"
		      "    stw 0[%1] = %0 \n"
		      "    and $r0.1 = $r0.1,~1\n":"=&r" (tmp)
		      :"r"(v)
		      :"r0.49", "memory");

	return tmp;
}

static inline int atomic_dec_return(atomic_t * v)
{
	int tmp;

	asm volatile ("1 : ldw %0   = 0[%1] \n"
		      "    mov $r0.49 = 1b \n"
		      "    or  $r0.1 = $r0.1,1 \n"
		      "    ;; \n"
		      "    add %0 = %0,-1 \n"
		      "    ;; \n"
		      "    stw 0[%1] = %0 \n"
		      "    and $r0.1 = $r0.1,~1\n":"=&r" (tmp)
		      :"r"(v)
		      :"r0.49", "memory");

	return tmp;
}

/**
 * atomic_sub_and_test - subtract value from variable and test result
 * @i: integer value to subtract
 * @v: pointer of type atomic_t
 *
 * Atomically subtracts @i from @v and returns
 * true if the result is zero, or false for all
 * other cases.
 */

#define atomic_sub_and_test(i,v) (atomic_sub_return((i), (v)) == 0)

/**
 * atomic_inc - increment atomic variable
 * @v: pointer of type atomic_t
 *
 * Atomically increments @v by 1
 */

static inline void atomic_inc(atomic_t * v)
{
	int tmp;

	asm volatile ("1 : ldw %0   = 0[%1] \n"
		      "    mov $r0.49 = 1b \n"
		      "    or  $r0.1 = $r0.1,1 \n"
		      "    ;; \n"
		      "    add %0 = %0,1 \n"
		      "    ;; \n"
		      "    stw 0[%1] = %0 \n"
		      "    and $r0.1 = $r0.1,~1":"=&r" (tmp)
		      :"r"(v)
		      :"r0.49", "memory");
}

/**
 * atomic_dec - decrement atomic variable
 * @v: pointer of type atomic_t
 *
 * Atomically decrements @v by 1.
 */

static inline void atomic_dec(atomic_t * v)
{
	int tmp;

	asm volatile ("1 : ldw %0   = 0[%1] \n"
		      "    mov $r0.49 = 1b \n"
		      "    or  $r0.1 = $r0.1,1 \n"
		      "    ;; \n"
		      "    add %0 = %0,-1 \n"
		      "    ;; \n"
		      "    stw 0[%1] = %0 \n"
		      "    and $r0.1 = $r0.1,~1":"=&r" (tmp)
		      :"r"(v)
		      :"r0.49", "memory");
}

/**
 * atomic_dec_and_test - decrement and test
 * @v: pointer of type atomic_t
 *
 * Atomically decrements @v by 1 and
 * returns true if the result is 0, or false for all other
 * cases.
 */

#define atomic_dec_and_test(v) (atomic_dec_return(v) == 0)

/**
 * atomic_inc_and_test - increment and test
 * @v: pointer of type atomic_t
 *
 * Atomically increments @v by 1
 * and returns true if the result is zero, or false for all
 * other cases.
 */
#define atomic_inc_and_test(v) (atomic_inc_return(v) == 0)

/**
 * atomic_add_negative - add and test if negative
 * @v: pointer of type atomic_t
 * @i: integer value to add
 *
 * Atomically adds @i to @v and returns true
 * if the result is negative, or false when
 * result is greater than or equal to zero.
 */
#define atomic_add_negative(i, v) (atomic_add_return(i, (v)) < 0)

#define atomic_cmpxchg(v, old, new) ((int)cmpxchg(&((v)->counter), old, new))
#define atomic_xchg(v, new) (xchg(&((v)->counter), new))

/**
 * atomic_add_unless - add unless the number is a given value
 * @v: pointer of type atomic_t
 * @a: the amount to add to v...
 * @u: ...unless v is equal to u.
 *
 * Atomically adds @a to @v, so long as it was not @u.
 * Returns non-zero if @v was not @u, and zero otherwise.
 */
static inline int atomic_add_unless(atomic_t * v, int a, int u)
{
	int val, res;

	asm volatile ("1 : ldw %0   = 0[%2] \n" "    mov $r0.49 = 1b \n" "    or  $r0.1 = $r0.1,1 \n" "    ;; \n" "    cmpne $b0.7 = %0,%4 \n" "	  cmpne %1 = %0,%4 \n" "    add %0 = %0,%3 \n" "    ;; \n" "	  brf $b0.7,2f\n" "    ;; \n" "    stw 0[%2] = %0 \n" "    and $r0.1 = $r0.1,~1 \n" "    ;; \n" "2:  and $r0.1 = $r0.1,~1 ":"=&r" (val) /* %0 */ , "=&r"(res)	/* %1 */
		      :"r"(v) /* %2 */ , "r"(a) /* %3 */ , "r"(u)	/* %4 */
		      :"r0.49", "memory", "b0.7");

	return res;
}

#define atomic_inc_not_zero(v) atomic_add_unless((v), 1, 0)

static inline void atomic_clear_mask(unsigned int mask, atomic_t * v)
{
	int tmp;

	asm volatile ("1 : ldw %0   = 0[%1] \n"
		      "    mov $r0.49 = 1b \n"
		      "    or  $r0.1 = $r0.1,1 \n"
		      "    ;; \n"
		      "    and %0 = %0,%2 \n"
		      "    ;; \n"
		      "    stw 0[%1] = %0 \n"
		      "    and $r0.1 = $r0.1,~1":"=&r" (tmp)
		      :"r"(v), "r"(~mask)
		      :"r0.49", "memory");
}

static inline void atomic_set_mask(unsigned int mask, atomic_t * v)
{
	int tmp;

	asm volatile ("1 : ldw %0   = 0[%1] \n"
		      "    mov $r0.49 = 1b \n"
		      "    or  $r0.1 = $r0.1,1 \n"
		      "    ;; \n"
		      "    or %0 = %0,%2 \n"
		      "    ;; \n"
		      "    stw 0[%1] = %0 \n"
		      "    and $r0.1 = $r0.1,~1":"=&r" (tmp)
		      :"r"(v), "r"(mask)
		      :"r0.49", "memory");
}

/* Atomic operations are already serializing */
#define smp_mb__before_atomic_dec()	barrier()
#define smp_mb__after_atomic_dec()	barrier()
#define smp_mb__before_atomic_inc()	barrier()
#define smp_mb__after_atomic_inc()	barrier()

#include <asm-generic/atomic-long.h>

#endif /* _ASM_ST200_ATOMIC_H */
