/*
 *  include/asm-st200/bitops.h
 *
 *  Copyright (C) 2002 STMicroelectronics Limited
 *	Author: David McKay <david.mckay@st.com>
 *
 */

#ifndef _ASM_ST200_BITOPS_H
#define _ASM_ST200_BITOPS_H

#ifndef _LINUX_BITOPS_H
#error only <linux/bitops.h> can be included directly
#endif

#ifdef __KERNEL__

/*
 * All bit operations return 0 if the bit was cleared before the
 * operation and != 0 if it was not.
 *
 * bit 0 is the LSB of addr; bit 32 is the LSB of (addr+1).
 */

/**
 * set_bit - Atomically set a bit in memory
 * @nr: the bit to set
 * @addr: the address to start counting from
 *
 * This function is atomic and may not be reordered.  See __set_bit()
 * if you do not require the atomic guarantees.
 * Note that @nr may be almost arbitrarily large; this function is not
 * restricted to acting on a single-word quantity.
 */
static inline void set_bit(int nr, volatile unsigned long *addr)
{
	int mask;
	volatile unsigned long *a = addr;
	unsigned long tmp;

	a += nr >> 5;
	mask = 1 << (nr & 0x1f);

	asm volatile ("1 : ldw %0   = 0[%1] \n"
		      "    mov $r62 = 1b \n"
		      "    or  $r12 = $r12,1 \n"
		      "    ;; \n"
		      "    or %0 = %0,%2 \n"
		      "    ;; \n"
		      "    stw 0[%1] = %0 \n"
		      "    and $r12 = $r12,~1\n":"=&r" (tmp)
		      :"r"(a), "r"(mask)
		      :"r62", "memory");
}

/**
 * clear_bit - Clears a bit in memory
 * @nr: Bit to clear
 * @addr: Address to start counting from
 *
 * clear_bit() is atomic and may not be reordered.  However, it does
 * not contain a memory barrier, so if it is used for locking purposes,
 * you should call smp_mb__before_clear_bit() and/or smp_mb__after_clear_bit()
 * in order to ensure changes are visible on other processors.
 */
static inline void clear_bit(int nr, volatile unsigned long *addr)
{
	int mask;
	volatile unsigned long *a = addr;
	unsigned long tmp;

	a += nr >> 5;
	mask = ~(1 << (nr & 0x1f));

	asm volatile ("1 : ldw %0   = 0[%1] \n"
		      "    mov $r62 = 1b \n"
		      "    or  $r12 = $r12,1 \n"
		      "    ;; \n"
		      "    and %0 = %0,%2 \n"
		      "    ;; \n"
		      "    stw 0[%1] = %0 \n"
		      "    and $r12 = $r12,~1\n":"=&r" (tmp)
		      :"r"(a), "r"(mask)
		      :"r62", "memory");
}

#define smp_mb__before_clear_bit()	barrier()
#define smp_mb__after_clear_bit()	barrier()

/**
 * change_bit - Toggle a bit in memory
 * @nr: Bit to change
 * @addr: Address to start counting from
 *
 * change_bit() is atomic and may not be reordered.
 * Note that @nr may be almost arbitrarily large; this function is not
 * restricted to acting on a single-word quantity.
 */
static inline void change_bit(int nr, volatile unsigned long *addr)
{
	int mask;
	volatile unsigned long *a = addr;
	unsigned long tmp;

	a += nr >> 5;
	mask = 1 << (nr & 0x1f);
	asm volatile ("1 : ldw %0   = 0[%1] \n"
		      "    mov $r62 = 1b \n"
		      "    or  $r12 = $r12,1 \n"
		      "    ;; \n"
		      "    xor %0 = %0,%2 \n"
		      "    ;; \n"
		      "    stw 0[%1] = %0 \n"
		      "    and $r12 = $r12,~1\n":"=&r" (tmp)
		      :"r"(a), "r"(mask)
		      :"r62", "memory");

}

/**
 * test_and_set_bit - Set a bit and return its old value
 * @nr: Bit to set
 * @addr: Address to count from
 *
 * This operation is atomic and cannot be reordered.
 * It also implies a memory barrier.
 */
static inline int test_and_set_bit(int nr, volatile unsigned long *addr)
{
	int mask, retval;
	volatile unsigned long *a = addr;
	unsigned long tmp;

	a += nr >> 5;
	mask = 1 << (nr & 0x1f);

	asm volatile ("1 : ldw %0   = 0[%2] \n"
		      "    mov $r62 = 1b \n"
		      "    or  $r12 = $r12,1 \n"
		      "    ;; \n"
		      "    and %1 = %0,%3\n"
		      "    or %0 = %0,%3 \n"
		      "    ;; \n"
		      "    cmpne %1 = %1,0\n"
		      "    stw 0[%2] = %0 \n"
		      "    and $r12 = $r12,~1\n":"=&r" (tmp), "=&r"(retval)
		      :"r"(a), "r"(mask)
		      :"r62", "memory");

	return retval;
}

/**
 * test_and_clear_bit - Clear a bit and return its old value
 * @nr: Bit to clear
 * @addr: Address to count from
 *
 * This operation is atomic and cannot be reordered.
 * It also implies a memory barrier.
 */
static inline int test_and_clear_bit(int nr, volatile unsigned long *addr)
{
	int mask, retval;
	volatile unsigned long *a = addr;
	unsigned long tmp;

	a += nr >> 5;
	mask = 1 << (nr & 0x1f);

	asm volatile ("1 : ldw %0   = 0[%2] \n"
		      "    mov $r62 = 1b \n"
		      "    or  $r12 = $r12,1 \n"
		      "    ;; \n"
		      "    and %1 = %0,%3\n"
		      "    and %0 = %0,%4 \n"
		      "    ;; \n"
		      "    cmpne %1 = %1,0\n"
		      "    stw 0[%2] = %0 \n"
		      "    and $r12 = $r12,~1\n":"=&r" (tmp), "=&r"(retval)
		      :"r"(a), "r"(mask), "r"(~mask)
		      :"r62", "memory");

	return retval;
}

/**
 * test_and_change_bit - Change a bit and return its old value
 * @nr: Bit to change
 * @addr: Address to count from
 *
 * This operation is atomic and cannot be reordered.
 * It also implies a memory barrier.
 */
static __inline__ int test_and_change_bit(int nr, volatile unsigned long *addr)
{
	int mask, retval;
	volatile unsigned long *a = addr;
	unsigned long tmp;

	a += nr >> 5;
	mask = 1 << (nr & 0x1f);

	asm volatile ("1 : ldw %0   = 0[%2] \n"
		      "    mov $r62 = 1b \n"
		      "    or  $r12 = $r12,1 \n"
		      "    ;; \n"
		      "    and %1 = %0,%3\n"
		      "    xor %0 = %0,%3 \n"
		      "    ;; \n"
		      "    cmpne %1 = %1,0\n"
		      "    stw 0[%2] = %0 \n"
		      "    and $r12 = $r12,~1\n":"=&r" (tmp), "=&r"(retval)
		      :"r"(a), "r"(mask)
		      :"r62", "memory");

	return retval;
}

#include <linux/compiler.h>

#include <asm-generic/bitops/non-atomic.h>
#include <asm-generic/bitops/find.h>
#include <asm-generic/bitops/ffz.h>
#include <asm-generic/bitops/ffs.h>
#include <asm-generic/bitops/__ffs.h>
#include <asm-generic/bitops/fls.h>
#include <asm-generic/bitops/__fls.h>
#include <asm-generic/bitops/fls64.h>
#include <asm-generic/bitops/hweight.h>
#include <asm-generic/bitops/sched.h>
#include <asm-generic/bitops/ext2-non-atomic.h>
#include <asm-generic/bitops/ext2-atomic.h>
#include <asm-generic/bitops/minix.h>
#include <asm-generic/bitops/lock.h>

#endif /* __KERNEL__ */

#endif /* _ASM_ST200_BITOPS_H */
