#ifndef _ASM_ST200_DMA_MAPPING_H
#define _ASM_ST200_DMA_MAPPING_H

#include <linux/mm.h>
#include <linux/scatterlist.h>
#include <asm/cache.h>
#include <asm/io.h>

/* These should really allocate some cached memory */
#define dma_alloc_noncoherent(d, s, h, f) dma_alloc_coherent(d, s, h, f)
#define dma_free_noncoherent(d, s, v, h) dma_free_coherent(d, s, v, h)

void *dma_alloc_coherent(struct device *dev, size_t size,
			 dma_addr_t *dma_handle, gfp_t flag);

void dma_free_coherent(struct device *dev, size_t size,
		       void *vaddr, dma_addr_t dma_handle);

extern void st200_flush_dcache_range(unsigned long addr, unsigned long size);

/* Nobody should put anything here on an ST200 system */
#define DMA_ERROR_CODE          (~(dma_addr_t)0x0)

static inline int dma_mapping_error(struct device *dev, dma_addr_t dma_addr)
{
	return (dma_addr == DMA_ERROR_CODE);
}

static inline void
dma_cache_sync(struct device *dev, void *vaddr, size_t size,
	       enum dma_data_direction direction)
{
	/* For ST240 we can optimise this, as we now have invalidate */
	st200_flush_dcache_range((unsigned long)vaddr, (unsigned long)size);
}

static inline dma_addr_t
dma_map_single(struct device *dev, void *ptr, size_t size,
	       enum dma_data_direction direction)
{
	BUG_ON(direction == DMA_NONE);

	dma_cache_sync(dev, ptr, size, direction);
	return virt_to_phys(ptr);
}

static inline void
dma_unmap_single(struct device *dev, dma_addr_t dma_addr, size_t size,
		 enum dma_data_direction direction)
{
	BUG_ON(direction == DMA_NONE);
}

static inline int
dma_map_sg(struct device *dev, struct scatterlist *sg, int nents,
	   enum dma_data_direction direction)
{
	int i;
	void set_watchpoint(void);

	BUG_ON(direction == DMA_NONE);

	for (i = 0; i < nents; i++) {
		dma_cache_sync(dev, sg_virt(sg + i), sg[i].length, direction);
		sg[i].dma_address = sg_phys(sg + i);
	}

	return nents;
}

static inline dma_addr_t
dma_map_page(struct device *dev, struct page *page, unsigned long offset,
	     size_t size, enum dma_data_direction direction)
{
	return dma_map_single(dev, page_address(page) + offset, size,
			      direction);
}

static inline void
dma_unmap_page(struct device *dev, dma_addr_t dma_handle, size_t size,
	       enum dma_data_direction direction)
{
	dma_unmap_single(dev, dma_handle, size, direction);
}

static inline void
dma_unmap_sg(struct device *dev, struct scatterlist *sg, int nhwentries,
	     enum dma_data_direction direction)
{
	BUG_ON(direction == DMA_NONE);
}

static inline void
dma_sync_single_for_cpu(struct device *dev, dma_addr_t dma_handle, size_t size,
			enum dma_data_direction direction)
{
	dma_cache_sync(dev, (void *)(phys_to_virt(dma_handle)), size,
		       direction);
}

static inline void
dma_sync_single_for_device(struct device *dev, dma_addr_t dma_handle,
			   size_t size, enum dma_data_direction direction)
{
	dma_cache_sync(dev, (void *)(phys_to_virt(dma_handle)), size,
		       direction);
}

static inline void
dma_sync_single_range_for_cpu(struct device *dev, dma_addr_t dma_handle,
			      unsigned long offset, size_t size,
			      enum dma_data_direction direction)
{
	dma_sync_single_for_cpu(dev, dma_handle + offset, size, direction);
}

static inline void
dma_sync_single_range_for_device(struct device *dev, dma_addr_t dma_handle,
				 unsigned long offset, size_t size,
				 enum dma_data_direction direction)
{
	dma_sync_single_for_device(dev, dma_handle + offset, size, direction);
}

static inline void
dma_sync_sg_for_cpu(struct device *dev, struct scatterlist *sg, int nelems,
		    enum dma_data_direction direction)
{
	int i;

	for (i = 0; i < nelems; i++) {
		dma_cache_sync(dev, sg_virt(sg + i), sg[i].length, direction);
	}
}

static inline void
dma_sync_sg_for_device(struct device *dev, struct scatterlist *sg, int nelems,
		       enum dma_data_direction direction)
{
	dma_sync_sg_for_cpu(dev, sg, nelems, direction);
}

static inline int dma_supported(struct device *dev, u64 mask)
{
	return 1;
}

static inline int dma_set_mask(struct device *dev, u64 mask)
{
	if (!dev->dma_mask || !dma_supported(dev, mask))
		return -EIO;

	*dev->dma_mask = mask;

	return 0;
}

static inline int dma_get_cache_alignment(void)
{
	/* Return the maximum possible, to be safe */
	return L1_CACHE_BYTES;
}

/* This is wrong, but the API is crap, you really need the virtual address to figure out
 * if the particular area you have been given is coherent or not. There is no easy way to
 * figure it out given just the physical address.
 */
#define dma_is_consistent(d,h)	(0)

#endif /*_ASM_ST200_DMA_MAPPING_H */
