/*
 * include/asm-st200/pgtable-mmu.h
 *
 * Copyright (C) 2002-3 STMicroelectronics Limited
 *	Author: Stuart Menefy <stuart.menefy@st.com>
 *
 */

#ifndef _ASM_ST200_PGTABLE_MMU_H
#define _ASM_ST200_PGTABLE_MMU_H

#include <asm-generic/4level-fixup.h>

#include <asm/page.h>


#ifndef __ASSEMBLY__
struct mm_struct;
struct vm_area_struct;
#endif

/*
 * The Linux memory management assumes a three-level page table setup. On
 * the ST200, we use that, but "fold" the mid level into the top-level page
 * table.
 *
 * Given a virtual address:
 * - the lowest PAGE_SHIFT bits determine offset into the page
 * - the next higher PAGE_SHIFT-2 bits determine the pte# in the pagetable
 *   directory (the -2 is from the 4 byte ptes, and each page table is a
 *   single page long).
 * - the remaining bits determine pgd#.
 *
 * Pictorially (assuming an 8K page):
 *
 *        pgd#                  pte#                page offset
 * +----------------+----------------------+--------------------------+
 * |31            24|23                  13|12                       0|
 * +----------------+----------------------+--------------------------+
 */

/* PMD_SHIFT determines the size of the area a second-level page table can map */
#define PMD_SHIFT       (PAGE_SHIFT + (PAGE_SHIFT-2))
#define PMD_SIZE        (1UL << PMD_SHIFT)
#define PMD_MASK        (~(PMD_SIZE-1))

/* PGDIR_SHIFT determines what a third-level page table entry can map */
#define PGDIR_SHIFT	PMD_SHIFT
#define PGDIR_SIZE	(1UL << PGDIR_SHIFT)
#define PGDIR_MASK	(~(PGDIR_SIZE-1))

/* Entries per page directory level: we use two-level, so
 * we don't really have any PMD directory physically.
 */
#define PTRS_PER_PTE    (PAGE_SIZE >> 2)
#define PTRS_PER_PMD    1
#define PTRS_PER_PGD    (1<<(32 - PGDIR_SHIFT))

#define USER_PTRS_PER_PGD       (TASK_SIZE/PGDIR_SIZE)
#define FIRST_USER_ADDRESS      0

#ifdef UNKNOWN
#define USER_PGD_PTRS (PAGE_OFFSET >> PGDIR_SHIFT)
#define KERNEL_PGD_PTRS (PTRS_PER_PGD-USER_PGD_PTRS)
#endif

#define pte_ERROR(e) \
	printk("%s:%d: bad pte %08lx.\n", __FILE__, __LINE__, pte_val(e))
#define pmd_ERROR(e) \
	printk("%s:%d: bad pmd %08lx.\n", __FILE__, __LINE__, pmd_val(e))
#define pgd_ERROR(e) \
	printk("%s:%d: bad pgd %08lx.\n", __FILE__, __LINE__, pgd_val(e))

/*
 * Align vmalloc and boot time mappings on a 16M boundary, to try and make
 * best use of pte directories.
 * We also leave a gap between physical memory, vmalloc, boot mappings
 * and control registers to try and catch any out-of-bounds memory accesses.
 * The allocation routines leave a page sized hole between each allocated
 * area for the same reason. ;)
 */
#define VMALLOC_OFFSET	(16*1024*1024)
#define VMALLOC_START	(0xc0000000 & ~(VMALLOC_OFFSET - 1))
#define VMALLOC_END	((CONTROL_START - (3 * VMALLOC_OFFSET)) \
				& ~(VMALLOC_OFFSET - 1))

/* Area of virtual memory reserved for dma_coherent allocations */
#define CONSISTENT_START	((CONTROL_START - (2 * VMALLOC_OFFSET)) \
					& ~(VMALLOC_OFFSET - 1))
#define CONSISTENT_END		((CONTROL_START - VMALLOC_OFFSET) \
					& ~(VMALLOC_OFFSET - 1))

#define VMALLOC_VMADDR(x) ((unsigned long)(x))

/* TLB_ENTRY2 */
#define _PAGE_PADDR_MASK	0x0007ffff

/* TLB_ENTRY0 */
#define _PAGE_BIT_SHARED	19

#define _PAGE_SHARED		(1 << _PAGE_BIT_SHARED)

/* Software flags */
#define _PAGE_BIT_ACCESSED	20
#define _PAGE_BIT_SOFT_USER_X	21

#define _PAGE_ACCESSED		(1 << _PAGE_BIT_ACCESSED)
#define _PAGE_SOFT_USER_X	(1 << _PAGE_BIT_SOFT_USER_X)

/* TLB_ENTRY0 */
#define _PAGE_BIT_PROT_SUPER_W	22
#define _PAGE_BIT_PROT_USER_X	23
#define _PAGE_BIT_PROT_USER_R	24
#define _PAGE_BIT_PROT_USER_W	25
#define _PAGE_BIT_DIRTY		26
#define _PAGE_BIT_POLICY0	27
#define _PAGE_BIT_POLICY1	28
#define _PAGE_BIT_UNUSED	29	/* when pte_present */
#define _PAGE_BIT_FILE		29	/* when !pte_present */
#define _PAGE_BIT_SIZE0		30
#define _PAGE_BIT_SIZE1		31

#define _PAGE_PROT_SUPER_W	(1 << _PAGE_BIT_PROT_SUPER_W)
#define _PAGE_PROT_USER_X	(1 << _PAGE_BIT_PROT_USER_X)
#define _PAGE_PROT_USER_R	(1 << _PAGE_BIT_PROT_USER_R)
#define _PAGE_PROT_USER_W	(1 << _PAGE_BIT_PROT_USER_W)
#define _PAGE_PROT_ALL_W        ( _PAGE_PROT_SUPER_W |  _PAGE_PROT_USER_W )
#define _PAGE_DIRTY		(1 << _PAGE_BIT_DIRTY)
#define _PAGE_POLICY_NC		(0 << _PAGE_BIT_POLICY0)
#define _PAGE_POLICY_C		(1 << _PAGE_BIT_POLICY0)
#define _PAGE_POLICY_NCWC	(2 << _PAGE_BIT_POLICY0)
#define _PAGE_POLICY_MASK	(3 << _PAGE_BIT_POLICY0)
#define _PAGE_FILE		(1 << _PAGE_BIT_FILE)
#define _PAGE_SIZE_DISABLED	(0 << _PAGE_BIT_SIZE0)
#define _PAGE_SIZE_8K		(1 << _PAGE_BIT_SIZE0)
#define _PAGE_SIZE_4M		(2 << _PAGE_BIT_SIZE0)
#define _PAGE_SIZE_256M		(3 << _PAGE_BIT_SIZE0)
#define _PAGE_SIZE_MASK		(3 << _PAGE_BIT_SIZE0)

/* Masks and shifts to convert pte's to TLB entries */
#define _PAGE_TLB_ENTRY2_MASK		_PAGE_PADDR_MASK
#define _PAGE_TLB_ENTRY0_0_MASK		0x1Fc80000
#define _PAGE_TLB_ENTRY0_0_SHIFT	11
#define _PAGE_TLB_ENTRY0_1_MASK		0xc0000000
#define _PAGE_TLB_ENTRY0_1_SHIFT	10

/* Extra bits which need to be ORed into TLB_ENTRY0 */
#define TLB_ENTRY0_EXTRABITS	( TLB_ENTRY0_PROT_SUPER_X |  TLB_ENTRY0_PROT_SUPER_R )

/* Bits which are not modified when changing a pte's permissions */
#define _PAGE_CHG_MASK	(_PAGE_PADDR_MASK | _PAGE_ACCESSED | _PAGE_DIRTY)

/* Bits which are set for a 'normal' page */
#define _PAGE_NORMAL		(_PAGE_SIZE_8K | _PAGE_ACCESSED | _PAGE_POLICY_C)
#define _PAGE_NORMAL_NC		((_PAGE_NORMAL & ~_PAGE_POLICY_MASK) | _PAGE_POLICY_NC)
#define _PAGE_NORMAL_NCWC	((_PAGE_NORMAL & ~_PAGE_POLICY_MASK) | _PAGE_POLICY_NCWC)

#ifndef __ASSEMBLY__

/*
 * All the normal masks have the "page accessed" bits on, as any time
 * they are used, the page is accessed. They are cleared only by the
 * page-out routines.
 *
 * PAGE_COPY is used for the stack, so usually has the execute bit set.
 * However we are trying to avoid this, so don't for now, and see
 * what breaks.
 *
 * Note we use _PAGE_SIZE_DISABLED to mark pages which are not present,
 * so we set _PAGE_SIZE_8K even for PAGE_NONE.
 */
#define PAGE_NONE	__pgprot(_PAGE_NORMAL)
#define PAGE_SHARED	__pgprot(_PAGE_NORMAL | _PAGE_PROT_USER_R | _PAGE_PROT_ALL_W)
#define PAGE_COPY	__pgprot(_PAGE_NORMAL | _PAGE_PROT_USER_R)
#define PAGE_READONLY	__pgprot(_PAGE_NORMAL | _PAGE_PROT_USER_R)
#define PAGE_KERNEL	__pgprot(_PAGE_NORMAL | _PAGE_SHARED | _PAGE_DIRTY | _PAGE_PROT_SUPER_W)
#define PAGE_KERNEL_NC	__pgprot(_PAGE_NORMAL_NC | _PAGE_SHARED | _PAGE_DIRTY | _PAGE_PROT_SUPER_W)
#define PAGE_KERNEL_NCWC	__pgprot(_PAGE_NORMAL_NCWC | _PAGE_SHARED | _PAGE_DIRTY | _PAGE_PROT_SUPER_W)

#define _PAGE_P(x) __pgprot(_PAGE_NORMAL | ((x) & (~_PAGE_PROT_ALL_W)))
#define _PAGE_S(x) __pgprot(_PAGE_NORMAL | (x))

	/* xwr */
#define __P000	_PAGE_P(0)
#define __P001	_PAGE_P(_PAGE_PROT_USER_R)
#define __P010	_PAGE_P(_PAGE_PROT_USER_W)
#define __P011	_PAGE_P(_PAGE_PROT_USER_W | _PAGE_PROT_USER_R)
#define __P100	_PAGE_P(_PAGE_SOFT_USER_X)
#define __P101	_PAGE_P(_PAGE_SOFT_USER_X | _PAGE_PROT_USER_R)
#define __P110	_PAGE_P(_PAGE_SOFT_USER_X | _PAGE_PROT_ALL_W)
#define __P111	_PAGE_P(_PAGE_SOFT_USER_X | _PAGE_PROT_ALL_W | _PAGE_PROT_USER_R)

#define __S000	_PAGE_S(0)
#define __S001	_PAGE_S(_PAGE_PROT_USER_R)
#define __S010	_PAGE_S(_PAGE_PROT_USER_W)
#define __S011	_PAGE_S(_PAGE_PROT_USER_W | _PAGE_PROT_USER_R)
#define __S100	_PAGE_S(_PAGE_SOFT_USER_X)
#define __S101	_PAGE_S(_PAGE_SOFT_USER_X | _PAGE_PROT_USER_R)
#define __S110	_PAGE_S(_PAGE_SOFT_USER_X | _PAGE_PROT_ALL_W)
#define __S111	_PAGE_S(_PAGE_SOFT_USER_X | _PAGE_PROT_ALL_W | _PAGE_PROT_USER_R)

/*
 * Certain architectures need to do special things when PTEs
 * within a page table are directly modified.  Thus, the following
 * hook is made available.
 *
 * pmds are folded into pgds so set_pmd doesnt get actually called,
 * but the define is needed for a generic inline function.
 */
void set_pte(pte_t * ptep, pte_t pte);

#define set_pte_at(mm,addr,ptep,pteval) set_pte(ptep,pteval)

#define set_pmd(pmdptr, pmdval) (*(pmdptr) = pmdval)
#define set_pgd(pgdptr, pgdval) (*(pgdptr) = pgdval)

static inline unsigned long pte_pfn(pte_t pte)
{
	return pte_val(pte) & _PAGE_PADDR_MASK;
}

static inline pte_t pfn_pte(unsigned long pfn, pgprot_t prot)
{
	return __pte(pfn | pgprot_val(prot));
}

static inline unsigned long pmd_pfn(pmd_t pmd)
{
	return __pa(pmd_val(pmd)) >> PAGE_SHIFT;
}

/* returns the "struct page *" corresponding to the pte/pmd */
#define pte_page(pte)		pfn_to_page(pte_pfn(pte))
#define pmd_page(pmd)		pfn_to_page(pmd_pfn(pmd))
#define pmd_pgtable(pmd) 	pmd_page(pmd)

#define mk_pte(page, pgprot)	pfn_pte(page_to_pfn(page), (pgprot))

/*
 * Remember pte_none() and pte_present() must work without checking
 * the other first. We use _PAGE_SIZE_DISABLED to mark pages not
 * present (and keep these bits free in swp_entry_t). pte_none will
 * always fail for present pages (because the _PAGE_SIZE bits will be
 * set), and for not present pages, because the offset cannot be zero.
 */
extern inline int pte_none(pte_t pte)
{
	return !pte_val(pte);
}

extern inline int pte_present(pte_t pte)
{
	return (pte_val(pte) & _PAGE_SIZE_MASK) != _PAGE_SIZE_DISABLED;
}

extern inline void pte_clear(struct mm_struct *mm, unsigned long addr,
			     pte_t * ptep)
{
	set_pte_at(mm, addr, ptep, __pte(0));
}

extern inline int pmd_none(pmd_t pmd)
{
	return !pmd_val(pmd);
}

extern void *high_memory;
extern inline int pmd_bad(pmd_t pmd)
{
	return (pmd_val(pmd) > (unsigned long)high_memory) ||
	    (pmd_val(pmd) < (PAGE_OFFSET + _MEMORY_OFFSET));
}

extern inline int pmd_present(pmd_t pmd)
{
	return pmd_val(pmd);
}

extern inline void pmd_clear(pmd_t * pmdp)
{
	set_pmd(pmdp, __pmd(0));
}

/*
 * The "pgd_xxx()" functions here are trivial for a folded two-level
 * setup: the pgd is never bad, and a pmd always exists (as it's folded
 * into the pgd entry)
 */
extern inline int pgd_none(pgd_t pgd)
{
	return 0;
}

extern inline int pgd_bad(pgd_t pgd)
{
	return 0;
}

extern inline int pgd_present(pgd_t pgd)
{
	return 1;
}

extern inline void pgd_clear(pgd_t * pgdp)
{
}

#define pages_to_mb(x) ((x) >> (20-PAGE_SHIFT))

/*
 * The following only work if pte_present() is true.
 * Undefined behaviour if not..
 */
static inline int pte_read(pte_t pte)
{
	return pte_val(pte) & _PAGE_PROT_USER_R;
}

static inline int pte_exec(pte_t pte)
{
	return pte_val(pte) & _PAGE_SOFT_USER_X;
}

static inline int pte_dirty(pte_t pte)
{
	return pte_val(pte) & _PAGE_DIRTY;
}

static inline int pte_young(pte_t pte)
{
	return pte_val(pte) & _PAGE_ACCESSED;
}

static inline int pte_write(pte_t pte)
{
	return pte_val(pte) & _PAGE_PROT_USER_W;
}

#define pte_special(pte) (0)

static inline pte_t pte_rdprotect(pte_t pte)
{
	pte_val(pte) &= ~_PAGE_PROT_USER_R;
	return pte;
}

static inline pte_t pte_exprotect(pte_t pte)
{
	pte_val(pte) &= ~(_PAGE_SOFT_USER_X | _PAGE_PROT_USER_X);
	return pte;
}

static inline pte_t pte_mkclean(pte_t pte)
{
	pte_val(pte) &= ~_PAGE_DIRTY;
	return pte;
}

static inline pte_t pte_mkold(pte_t pte)
{
	pte_val(pte) &= ~_PAGE_ACCESSED;
	return pte;
}

static inline pte_t pte_wrprotect(pte_t pte)
{
	pte_val(pte) &= ~_PAGE_PROT_ALL_W;
	return pte;
}

static inline pte_t pte_mkread(pte_t pte)
{
	pte_val(pte) |= _PAGE_PROT_USER_R;
	return pte;
}

static inline pte_t pte_mkexec(pte_t pte)
{
	pte_val(pte) |= _PAGE_SOFT_USER_X;
	return pte;
}

static inline pte_t pte_mkdirty(pte_t pte)
{
	pte_val(pte) |= _PAGE_DIRTY;
	return pte;
}

static inline pte_t pte_mkyoung(pte_t pte)
{
	pte_val(pte) |= _PAGE_ACCESSED;
	return pte;
}

static inline pte_t pte_mkwrite(pte_t pte)
{
	pte_val(pte) |= _PAGE_PROT_ALL_W;
	return pte;
}

static inline pte_t pte_mkspecial(pte_t pte)
{
	return pte;
}

static inline pte_t pte_modify(pte_t pte, pgprot_t newprot)
{
	return __pte((pte_val(pte) & _PAGE_CHG_MASK) | pgprot_val(newprot));
}

/* to find an entry in a page-table-directory. */
#define pgd_index(address) (( address >> PGDIR_SHIFT) & (PTRS_PER_PGD-1))
#define pgd_offset(mm, address) ((mm)->pgd+pgd_index(address))

/* to find an entry in a kernel page-table-directory */
#define pgd_offset_k(address) pgd_offset(&init_mm, address)

/* Find an entry in the second-level page table.. */
/* Trivial for the folded page table */
extern inline pmd_t *pmd_offset(pgd_t * dir, unsigned long address)
{
	return (pmd_t *) dir;
}

/* Find an entry in the third-level page table.. */
extern inline pte_t *pte_offset_kernel(pmd_t * pmdp, unsigned long address)
{
	return (pte_t *) pmd_val(*pmdp) +
	    ((address >> PAGE_SHIFT) & (PTRS_PER_PTE - 1));
}

#define pte_offset_map		pte_offset_kernel
#define pte_offset_map_nested	pte_offset_kernel
#define pte_unmap(pte)		do { } while (0)
#define pte_unmap_nested(pte)	do { } while (0)

typedef pte_t *pte_addr_t;

extern void update_mmu_cache(struct vm_area_struct *vma,
			     unsigned long address, pte_t pte);

/* Handling of pte entries which represent not_present() pages. */
/* We use _PAGE_SIZE_DISABLED (which is fortunatly zero) to flag pages
 * which are not present, so:
 *  bits 31-30	present flag (0)
 *  bit     29  file
 *  bits 28- 8	offset
 *  bits  7- 0	type
 */

/* Encode and de-code a swap entry */
#define __swp_type(x)			((x).val & 0xff)
#define __swp_offset(x)			(((x).val & 0x1fffffff) >> 8)
#define __swp_entry(type, offset)	((swp_entry_t) { ((type) | ((offset) << 8)) & 0x1fffffff })
#define __pte_to_swp_entry(pte)		((swp_entry_t) { pte_val(pte) })
#define __swp_entry_to_pte(x)		((pte_t) { (x).val })

/* File offset in PTE support. */
#define pte_file(pte)		(pte_val(pte) & _PAGE_FILE)
#define pte_to_pgoff(pte)	(pte_val(pte) & 0x1fffffff)
#define pgoff_to_pte(off)	(__pte((off) | _PAGE_FILE))
#define PTE_FILE_MAX_BITS	(29UL)	/* We reserve 3 bits in total */

#endif /* !__ASSEMBLY__ */

#ifndef __ASSEMBLY__

extern pgd_t swapper_pg_dir[PTRS_PER_PGD];
extern void paging_init(void);

/*
 * ZERO_PAGE is a global shared page that is always zero: used
 * for zero-mapped memory areas etc..
 */
extern unsigned long empty_zero_page[PAGE_SIZE / sizeof(unsigned long)];
#define ZERO_PAGE(vaddr) (virt_to_page(empty_zero_page))

#endif /* !__ASSEMBLY__ */

#ifndef CONFIG_DISCONTIGMEM
/*
 * kern_addr_valid(ADDR) tests if ADDR is pointing to valid kernel
 * memory.  For the return value to be meaningful, ADDR must be >=
 * PAGE_OFFSET.  This operation can be relatively expensive (e.g.,
 * require a hash-, or multi-level tree-lookup or something of that
 * sort) but it guarantees to return TRUE only if accessing the page
 * at that address does not cause an error.  Note that there may be
 * addresses for which kern_addr_valid() returns FALSE even though an
 * access would not cause an error (e.g., this is typically true for
 * memory mapped I/O regions.
 *
 * XXX Need to implement this for ST200.
 */
#define kern_addr_valid(addr)	(1)
#endif /* !CONFIG_DISCONTIGMEM */

#ifndef __ASSEMBLY__

/*
 * remap a physical address `phys' of size `size' with page protection `prot'
 * into virtual address `from'
 */
#define io_remap_page_range(vma,from,phys,size,prot) \
		remap_pfn_range(vma, from, (phys) >> PAGE_SHIFT, size, prot)

/*
 * Change the page cache policy.
 */
#define pgprot_noncached(prot)		\
	__pgprot((pgprot_val(prot) & ~_PAGE_POLICY_MASK) | _PAGE_POLICY_NC)
#define pgprot_writecombine(prot)	\
	__pgprot((pgprot_val(prot) & ~_PAGE_POLICY_MASK) | _PAGE_POLICY_NCWC)

#define pgprot_cached(prot)	\
	__pgprot((pgprot_val(prot) & ~_PAGE_POLICY_MASK) | _PAGE_POLICY_C)
#include <asm-generic/pgtable.h>

/*
 * remap a physical page `pfn' of size `size' with page protection `prot'
 * into virtual address `from'
 */
#define io_remap_pfn_range(vma,from,pfn,size,prot) \
                remap_pfn_range(vma, from, pfn, size, prot)

/*
 * No page table caches to initialise
 */
#define pgtable_cache_init() do { } while (0)

#endif /* __ASSEMBLY__ */

#endif /* _ASM_ST200_PGTABLE_MMU_H */
