/*
 * include/asm-st200/uaccess.h
 *
 * Copyright (C) 2002 STMicroelectronics Limited
 *	Author: Stuart Menefy <stuart.menefy@st.com>
 *
 */

#ifndef _ASM_ST200_UACCESS_H
#define _ASM_ST200_UACCESS_H

/*
 * User space memory access functions
 */
#include <linux/sched.h>
#include <asm/errno.h>

#define VERIFY_READ 0
#define VERIFY_WRITE 1

/*
 * The exception table consists of pairs of addresses: the first is the
 * address of an instruction that is allowed to fault, and the second is
 * the address at which the program should continue.  No registers are
 * modified, so it is entirely up to the continuation code to figure out
 * what to do.
 *
 * All the routines below use bits of fixup code that are out of line
 * with the main instruction path.  This means when everything is well,
 * we don't even have to jump over them.  Further, they do not intrude
 * on our cache or tlb entries.
 */

struct exception_table_entry {
	unsigned long insn, fixup;
};

/* Returns 0 if exception not found and fixup otherwise.  */
extern unsigned long search_exception_table(unsigned long);

/*
 * The fs value determines whether argument validity checking should be
 * performed or not.  If get_fs() == USER_DS, checking is performed, with
 * get_fs() == KERNEL_DS, checking is bypassed.
 *
 * For historical reasons (Data Segment Register?), these macros are misnamed.
 */

#define get_ds()	(KERNEL_DS)
#define get_fs()	(current_thread_info()->addr_limit)
#define set_fs(x)	(current_thread_info()->addr_limit = (x))

#define segment_eq(a,b)	((a).seg == (b).seg)

#define MAKE_MM_SEG(s)  ((mm_segment_t) { (s) })

/* addr_limit is the maximum accessible address for the task. we misuse
 * the KERNEL_DS and USER_DS values to both assign and compare the
 * addr_limit values through the equally misnamed get/set_fs macros.
 * (see above)
 */

#define KERNEL_DS	MAKE_MM_SEG(0xffffffff)
#define USER_DS		MAKE_MM_SEG(PAGE_OFFSET)

/*
 * Ideally we should be doing this using assembler:
 *   This needs 33-bit arithmetic. We have a carry..
 *
 *   sum := addr + size;  carry? --> result = false;
 *   if (sum >= addr_limit) result = false;
 *   otherwise result = true
 *
 * but we don't have inline asm, so hack it using long longs.
 */
static inline int access_ok(int type, const void *addr, unsigned long size)
{
	return ((unsigned long long)(unsigned long)addr +
		(unsigned long long)size) <=
	    (unsigned long long)current_thread_info()->addr_limit.seg;
}

static inline int verify_area(int type, const void *addr, unsigned long size)
{
	return access_ok(type, addr, size) ? 0 : -EFAULT;
}

/*
 * These are the main single-value transfer routines.  They automatically
 * use the right size if we just have the right pointer type.
 *
 * As the ST200 uses the same address space for kernel and user
 * data, we can just do these as direct assignments.  (Of course, the
 * exception handling means that it's no longer "just"...)
 *
 * Careful to not
 * (a) re-use the arguments for side effects (sizeof/typeof is ok)
 * (b) require any knowledge of processes at this stage
 *
 * These functions return 0 on success and -EFAULT on error.
 */
#define put_user(x,ptr) \
  __put_user_check((__typeof__(*(ptr)))(x),(ptr),sizeof(*(ptr)))
#define get_user(x,ptr) \
  __get_user_check((x),(ptr),sizeof(*(ptr)))

/*
 * The "__xxx" versions do not do address space checking, useful when
 * doing multiple accesses to the same area (the programmer has to do the
 * checks by hand with "access_ok()")
 */
#define __put_user(x,ptr) \
  __put_user_nocheck((__typeof__(*(ptr)))(x),(ptr),sizeof(*(ptr)))
#define __get_user(x,ptr) \
  __get_user_nocheck((x),(ptr),sizeof(*(ptr)))

extern long __get_user_asm_byte(const void *ptr, long *retval);
extern long __get_user_asm_half(const void *ptr, long *retval);
extern long __get_user_asm_word(const void *ptr, long *retval);
extern long long __get_user_asm_double(const void *ptr, long *retval);
extern long __get_user_bad(void);

extern void __put_user_asm_byte(long x, void *ptr, long *retval);
extern void __put_user_asm_half(long x, void *ptr, long *retval);
extern void __put_user_asm_word(long x, void *ptr, long *retval);
extern void __put_user_asm_double(long long x, void *ptr, long *retval);
extern void __put_user_bad(void);

#define __get_user_check(x,ptr,size)					\
({									\
	long __gu_err = -EFAULT;					\
	unsigned long __gu_val = 0;					\
	const __typeof__(*(ptr)) *__gu_addr = (ptr);			\
	if (access_ok(VERIFY_READ,__gu_addr,size)) {			\
		__gu_err = 0;						\
		__get_user_size(__gu_val,__gu_addr,(size),__gu_err);	\
	}								\
	(x) = (__typeof__(*(ptr)))__gu_val;				\
	__gu_err;							\
})

#define __get_user_nocheck(x,ptr,size)					\
({									\
	long __gu_err = 0;						\
	unsigned long __gu_val;						\
	__get_user_size(__gu_val,(ptr),(size),__gu_err);		\
	(x) = (__typeof__(*(ptr)))__gu_val;				\
	__gu_err;							\
})

#define __put_user_check(x,ptr,size)					\
({									\
	long __pu_err = -EFAULT;					\
	__typeof__(*(ptr)) *__pu_addr = (ptr);				\
	if (access_ok(VERIFY_WRITE,__pu_addr,size)) {			\
		__pu_err = 0;						\
		__put_user_size((x),__pu_addr,(size),__pu_err);		\
	}								\
	__pu_err;							\
})

#define __put_user_nocheck(x,ptr,size)					\
({									\
	long __pu_err = 0;						\
	__typeof__(*(ptr)) *__pu_addr = (ptr);				\
	__put_user_size((x),__pu_addr,(size),__pu_err);			\
	__pu_err;							\
})

#define __get_user_size(x,ptr,size,retval)				\
do {									\
	switch (size) {							\
	case 1:	(x) = __get_user_asm_byte(ptr,&retval);	break;		\
	case 2:	(x) = __get_user_asm_half(ptr,&retval);	break;		\
	case 4:	(x) = __get_user_asm_word(ptr,&retval);	break;		\
	case 8:	(x) = __get_user_asm_double(ptr,&retval); break;	\
	default: (x) = __get_user_bad();				\
	}								\
} while (0)

#define __put_user_size(x,ptr,size,retval)				\
do {									\
	switch (size) {							\
	case 1: __put_user_asm_byte(x,ptr,&retval);	break;	\
	case 2: __put_user_asm_half(x,ptr,&retval);	break;	\
	case 4: __put_user_asm_word(x,ptr,&retval);	break;	\
	case 8: __put_user_asm_double(x,ptr,&retval);	break;	\
	default: __put_user_bad();					\
	}								\
} while (0)

unsigned long __do_copy_user(void *to, const void *from, unsigned long n);
unsigned long __do_clear_user(void *to, unsigned long n);
long __do_strncpy_from_user(char *dst, const char *src, long count);
long __do_strnlen_user(const char *s, long n);

/*
 * All the {__,}copy_{to,from}_user and {__,}clear_user functions
 * return the number of bytes NOT copied.
 */
static __inline__ unsigned long copy_from_user(void *to, const void *from,
					       unsigned long n)
{
	if (access_ok(VERIFY_READ, from, n))
		n = __do_copy_user(to, from, n);
	else			/* security hole - plug it */
		memset(to, 0, n);
	return n;
}

static __inline__ unsigned long __copy_from_user_inatomic(void *to,
							  const void *from,
							  unsigned long n)
{
	return __do_copy_user(to, from, n);
}

static inline unsigned long __copy_from_user(void *to, const void __user * from,
					     unsigned long n)
{
	might_sleep();
	return __copy_from_user_inatomic(to, from, n);
}

static __inline__ unsigned long copy_to_user(void *to, const void *from,
					     unsigned long n)
{
	if (access_ok(VERIFY_WRITE, to, n))
		n = __do_copy_user(to, from, n);
	return n;
}

static __inline__ unsigned long __copy_to_user_inatomic(void *to,
							const void *from,
							unsigned long n)
{
	return __do_copy_user(to, from, n);
}

static inline unsigned long __copy_to_user(void __user * to, const void *from,
					   unsigned long n)
{
	might_sleep();
	return __copy_to_user_inatomic(to, from, n);
}

static __inline__ unsigned long clear_user(void *to, unsigned long n)
{
	if (access_ok(VERIFY_WRITE, to, n))
		n = __do_clear_user(to, n);
	return n;
}

static __inline__ unsigned long __clear_user(void *to, unsigned long n)
{
	return __do_clear_user(to, n);
}

/*
 * {__,}strcpy_from_user returns -EFAULT if exception, count if the entire
 * buffer filled, else strlen (without trailing nul).
 */
static __inline__ long strncpy_from_user(char *dst, const char *src, long count)
{
	long res = -EFAULT;
	if (access_ok(VERIFY_READ, src, 1))
		res = __do_strncpy_from_user(dst, src, count);
	return res;
}

static __inline__ long __strncpy_from_user(char *dst, const char *src,
					   long count)
{
	return __do_strncpy_from_user(dst, src, count);
}

/*
 * strnlen_user returns 0 if exception before NUL or reaching the
 * supplied limit (N), a value greater than N if the limit would be
 * exceeded, else strlen (including trailing nul).
 */
static inline long strnlen_user(const char *s, long n)
{
	unsigned long res = 0;

	if (access_ok(VERIFY_READ, s, 0))
		res = __do_strnlen_user(s, n);
	return res;
}

#define strlen_user(s)	strnlen_user(s, ~0UL >> 1)

#endif /* _ASM_ST200_UACCESS_H */
